package com.jeffknecht.dynadao.api;

import java.util.HashMap;
import java.util.List;

import org.springframework.dao.IncorrectResultSizeDataAccessException;

public interface DynaDao {

	/**
	 * Returns all instances of the supplied class.
	 * 
	 * @param clazz May not be null.  Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	public abstract <T> List<T> findAll(Class<T> clazz);

	/**
	 * Uses a NamedParameterJdbcTemplate to populate a list
	 * of objects of the specified type which match the supplied
	 * criteria.
	 * 
	 * The where clause should be expressed using the
	 * property names defined in the supplied clazz.  These
	 * properties should be surrounded by {curlybraces}
	 * so that they can be replaced by their proper field
	 * name.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @param where May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied where clause.
	 * @return
	 */
	public abstract <T> List<T> find(Class<T> clazz, final String where,
			HashMap<String, ?> params);

	/**
	 * Uses a NamedParameterJdbcTemplate to locate a single
	 * of object of the specified type which match the supplied
	 * criteria.
	 * 
	 * The where clause should be expressed using the
	 * property names defined in the supplied clazz.  These
	 * properties should be surrounded by {curlybraces}
	 * so that they can be replaced by their proper field
	 * name.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @param where May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied where clause.
	 * @return May return null if no matching records are found.
	 * @throws IncorrectResultSizeDataAccessException if more than one matching record is found
	 */
	public abstract <T> T findOne(Class<T> clazz, final String where,
			HashMap<String, ?> params) throws IncorrectResultSizeDataAccessException;

	/**
	 * Inserts the supplied object.  If an auto-increment column is
	 * specified in the object's class definition, it is updated with
	 * the value generated by the database.
	 * 
	 * @param object May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return The number of records updated.
	 */
	public abstract int insert(Object object);

	/**
	 * Updates the supplied object.  The object must have at least one column
	 * annotated as a primary key field.
	 * 
	 * @param object May not be null. Must be annotated with a Table and at least one Column definition must be marked as a primary key field.
	 * @return The number of records updated.
	 */
	public abstract int update(Object object);

}