package com.jeffknecht.dynadao.impl;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.sql.DataSource;

import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.BeanPropertySqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.support.GeneratedKeyHolder;

import com.jeffknecht.dynadao.annotation.Column;
import com.jeffknecht.dynadao.api.DynaDao;
import com.jeffknecht.dynadao.api.DynamicDaoDataAccessException;


/**
 * Dynamic data access object.  Provides the ability to 
 * query and updated database tables based on annotations in the
 * DTO class definition.
 * 
 * Note that the intention here is not to replace JPA or Hibernate.
 * These frameworks are very good at what they do, but introduce some
 * interesting complexities with respect to session management.  Rather,
 * the intention of this library is to simplify the 80% usage scenario
 * of simple CRUD operations against a single table.
 * 
 * @author jeff.knecht
 *
 */
public class DynaDaoImpl implements DynaDao {

	NamedParameterJdbcTemplate namedParameterjdbcTemplate;
	JdbcTemplate jdbcTemplate;
	
	public DynaDaoImpl(DataSource ds) {
		this.jdbcTemplate = new JdbcTemplate(ds);
		this.namedParameterjdbcTemplate = new NamedParameterJdbcTemplate(ds);
	}
	
	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#findAll(java.lang.Class)
	 */
	public <T> List<T> findAll(Class<T> clazz) {
		String sql = baseSelectSQL(clazz);
		return jdbcTemplate.query(sql, new DynaRowMapper<T>(clazz), (Object[]) null);
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#find(java.lang.Class, java.lang.String, java.util.HashMap)
	 */
	public <T> List<T> find(Class<T> clazz, final String where, HashMap<String, ?> params) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		Pattern p = Pattern.compile("\\{[^\\{\\}]*\\}");	// match any text surrounded by {curly braces}
															// except ignore curly braces inside of curly braces
		
		String targetWhere = where.trim();
		Matcher matcher = p.matcher(targetWhere);
		while (matcher.find()) {
			String toReplace = matcher.group();
			String fieldName = toReplace.substring(1, toReplace.length() - 1);
			targetWhere = targetWhere.replace(toReplace, resolveColumnName(dtoHelper, fieldName));
			matcher = p.matcher(targetWhere);
		}
		
		sql.append(" ");
		if (!targetWhere.startsWith("where ")) {
			sql.append("where ");
		}
		sql.append(targetWhere);
		return namedParameterjdbcTemplate.query(sql.toString(), params, new DynaRowMapper<T>(clazz));
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#find(java.lang.Class, java.lang.String, java.util.HashMap)
	 */
	public <T> T findOne(Class<T> clazz, final String where, HashMap<String, ?> params) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		Pattern p = Pattern.compile("\\{[^\\{\\}]*\\}");	// match any text surrounded by {curly braces}
															// except ignore curly braces inside of curly braces
		
		String targetWhere = where.trim();
		Matcher matcher = p.matcher(targetWhere);
		while (matcher.find()) {
			String toReplace = matcher.group();
			String fieldName = toReplace.substring(1, toReplace.length() - 1);
			targetWhere = targetWhere.replace(toReplace, resolveColumnName(dtoHelper, fieldName));
			matcher = p.matcher(targetWhere);
		}
		
		sql.append(" ");
		if (!targetWhere.startsWith("where ")) {
			sql.append("where ");
		}
		sql.append(targetWhere);
		try {
			return namedParameterjdbcTemplate.queryForObject(sql.toString(), params, new DynaRowMapper<T>(clazz));
		} catch (EmptyResultDataAccessException ignored) {
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#insert(java.lang.Object)
	 */
	public int insert(Object object) {
		DTOHelper dtoHelper = DTOHelper.get(object.getClass());
		String sql = insertSQL(object.getClass());
		if (dtoHelper.autoIncrementFieldName() == null) {
			return namedParameterjdbcTemplate.update(sql, new BeanPropertySqlParameterSource(object));
		} else {
			String fieldName = dtoHelper.autoIncrementFieldName();
			GeneratedKeyHolder keyHolder = new GeneratedKeyHolder();
			int returnValue = namedParameterjdbcTemplate.update(sql, new BeanPropertySqlParameterSource(object), keyHolder);
			Number key = keyHolder.getKey();
			Field field = dtoHelper.field(fieldName);
			boolean accessible = field.isAccessible();
			field.setAccessible(true);
			try {
				field.setInt(object, key.intValue());
			} catch (IllegalAccessException e) {
				throw new DynamicDaoDataAccessException("Illegal access on field " + field, e);
			}
			field.setAccessible(accessible);
			return returnValue;
		}
	}
	
	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#update(java.lang.Object)
	 */
	public int update(Object object) {
		DTOHelper dtoHelper = DTOHelper.get(object.getClass());
		Set<String> primaryKeyFieldNames = dtoHelper.primaryKeyFieldNames();
		if (primaryKeyFieldNames.size() == 0) {
			throw new DynamicDaoDataAccessException("Unable to uniquely identify object " + object + ", because it has no primary key fields");
		}
		StringBuilder sqlBuilder = new StringBuilder(baseUpdateSQL(object.getClass()));
		sqlBuilder.append(" where ");
		int pkCounter = 0;
		for(String fieldName : primaryKeyFieldNames) {
			if (pkCounter > 0) {
				sqlBuilder.append(" and ");
			}
			sqlBuilder.append(dtoHelper.columnName(fieldName)).append(" = :").append(fieldName);
			pkCounter++;
		}
		return namedParameterjdbcTemplate.update(sqlBuilder.toString(), new BeanPropertySqlParameterSource(object));
	}

	/**
	 * Generates a select statement for the supplied class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	private String baseSelectSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		return "select " + columnList(dtoHelper) + " from " + dtoHelper.table();
	}

	/**
	 * Generates a sql insert statement for the supplied DTO class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	private String insertSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		
		StringBuilder sb = new StringBuilder("insert into ");
		sb.append(dtoHelper.table());
		sb.append(" (");
		int columnCount = 0;
		
		StringBuilder fieldBuilder = new StringBuilder();
		for (String fieldName : dtoHelper.fieldNames()) {
			Column column = dtoHelper.column(fieldName);
			if (!column.autoIncrement()) {
				if (columnCount > 0) {
					sb.append(", ");
					fieldBuilder.append(", ");
				}
				sb.append(column.value());
				fieldBuilder.append(":").append(fieldName);
				columnCount++;
			}
		}
		sb.append(") values (").append(fieldBuilder.toString()).append(")");
		return sb.toString();
	}

	/**
	 * Generates an update statement without the WHERE clause
	 * for the supplied DTO class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	private String baseUpdateSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sb = new StringBuilder("update ");
		sb.append(dtoHelper.table());
		sb.append(" set ");
		int columnCount = 0;
		
		for (String fieldName : dtoHelper.fieldNames()) {
			Column column = dtoHelper.column(fieldName);
			if (!column.autoIncrement()) {
				if (columnCount > 0) {
					sb.append(", ");
				}
				sb.append(column.value()).append(" = :").append(fieldName);
			}
		}
		return sb.toString();
	}

	/**
	 * Generates a comma-delimited list of column names based
	 * on the columns provided by the dtoHelper.
	 * 
	 * @param dtoHelper
	 * @return
	 */
	private String columnList(DTOHelper dtoHelper) {
		StringBuilder sb = new StringBuilder();
		for (Column column : dtoHelper.columns()) {
			if (sb.length() > 0) {
				sb.append(", ");
			}
			sb.append(column.value());			
		}
		return sb.toString();
	}

	/**
	 * Translates a field name into a database column name.
	 * 
	 * @param dtoHelper
	 * @param fieldName
	 * @return
	 */
	private String resolveColumnName(DTOHelper dtoHelper, String fieldName) {
		return dtoHelper.columnName(fieldName);
	}
}
